import json
import os
from abc import ABC, abstractmethod

import numpy as np
from umap import UMAP

from .model import DataModel


def _validate_path(paths: dict):
    return all(os.path.exists(path) for path in paths.values())


def _load_npy_file(path):
    return np.load(path, mmap_mode=None) if os.path.exists(path) else None


def _load_json_file(path, key):
    if os.path.exists(path):
        with open(path, "r") as f:
            return json.load(f).get(key)
    return None


class BaseDataLoader(ABC):
    def __init__(self):
        self.base_path = None
        self._data = None
        self._label = None
        self._legend = None
        self._precomputed_knn = None

    @abstractmethod
    def load_raw_data(self):
        raise FileNotFoundError("Data not found")

    def drop_cache(self, paths):
        for path in paths:
            if os.path.exists(path):
                os.remove(path)

    def _get_paths(self):
        return {
            "data": os.path.join(self.base_path, "data.npy"),
            "label": os.path.join(self.base_path, "label.npy"),
            "legend": os.path.join(self.base_path, "legend.json"),
            "knn_indices": os.path.join(self.base_path, "knn_indices.npy"),
            "knn_dists": os.path.join(self.base_path, "knn_dists.npy"),
        }

    def load_data(self, get_knn: bool = True, n_neighbors: int = 15):
        paths = self._get_paths()
        knn_paths = {k: paths[k] for k in ["knn_indices", "knn_dists"]}

        if not os.path.exists(paths["data"]):
            self.load_raw_data()

        self._data = _load_npy_file(paths["data"])
        self._label = _load_npy_file(paths["label"])
        self._legend = _load_json_file(paths["legend"], "legend")

        if get_knn and not _validate_path(knn_paths):
            self._precomputed_knn = self.compute_knn(self._data, n_neighbors)
            self.save_data(paths["knn_indices"], self._precomputed_knn[0])
            self.save_data(paths["knn_indices"], self._precomputed_knn[1])

        elif get_knn:
            self._precomputed_knn = (
                _load_npy_file(knn_paths["knn_indices"]),
                _load_npy_file(knn_paths["knn_dists"]),
            )
        else:
            self._precomputed_knn = (None, None)

    def save_data(self, file_path, data):
        if not os.path.exists(file_path):
            np.save(file_path, data)

    def get_data(self) -> DataModel:
        result = {
            "data": self._data,
            "label": self._label,
            "legend": self._legend,
            "precomputed_knn": self._precomputed_knn,
        }

        return result

    def scale_data(self, data: np.ndarray):
        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler()
        return scaler.fit_transform(data)

    def compute_knn(self, X, n_neighbors: int = 15):
        if X.shape[0] < 4096:
            self._precomputed_knn = (None, None)
            return self._precomputed_knn

        X = self.scale_data(X)
        reducer = UMAP(n_neighbors=n_neighbors)
        _ = reducer.fit_transform(X)
        self._precomputed_knn = (reducer._knn_indices, reducer._knn_dists)

        return self._precomputed_knn



from dataclasses import asdict
import numpy as np

from benchmark.measure import measure_accuracy


from ghostumap import GhostUMAP2


# def get_data(data_name: str):
#     dl = DataLoader(data_name)
#     X, y, legend, precomputed_knn = dl.get_data().values()

#     return X, y, legend, precomputed_knn


def eval_accuracy(
    X: np.ndarray,
    hprams: dict,
    benchmark_type: str,
    n_iter: int = 10,
    precomputed_knn: tuple = (None, None),
):
    gu = GhostUMAP2(precomputed_knn=precomputed_knn)
    O, G, alive_ghosts = gu.fit_transform(X, benchmark=benchmark_type, **hprams)

    unstable_ghosts = gu.get_unstable_ghosts(
        distance=hprams.get("distance", 0.1),
        sensitivity=hprams.get("sensitivity", 1),
    )

    result = asdict(gu.get_results())

    f1, precision, recall = measure_accuracy(
        y_true=unstable_ghosts, y_pred=alive_ghosts
    )

    result["original_embedding"] = O
    result["ghost_embedding"] = G
    result["alive_ghosts"] = alive_ghosts
    result["unstable_ghosts"] = unstable_ghosts
    result["f1"] = f1
    result["precision"] = precision
    result["recall"] = recall

    return result


def eval_acc_dropping(X: np.ndarray, hprams: dict, n_iter=10):
    return eval_accuracy(X, hprams, "accuracy_dropping", n_iter)


def eval_acc_halving(X: np.ndarray, hprams: dict, n_iter=10):
    return eval_accuracy(X, hprams, "accuracy_halving", n_iter)


def eval_time_dropping():
    pass


def eval_time_halving():
    pass



import numpy as np


def measure_accuracy(y_true, y_pred):
    TP = np.sum(np.logical_and(y_true == 1, y_pred == 1))
    FP = np.sum(np.logical_and(y_true == 0, y_pred == 1))
    FN = np.sum(np.logical_and(y_true == 1, y_pred == 0))
    TN = np.sum(np.logical_and(y_true == 0, y_pred == 0))

    precision = TP / (TP + FP) if TP + FP > 0 else 0
    recall = TP / (TP + FN) if TP + FN > 0 else 0

    f1 = (
        2 * (precision * recall) / (precision + recall) if precision + recall > 0 else 0
    )

    return f1, precision, recall


__all__ = ["measure_accuracy"]



from pydantic import validate_call

from .data.model import TDataName, DataModel

from .data.mnistSeries import MnistSeriesLoader
from .data.celegans import CelegansLoader
from .data.uciml import UcimlLoader
from .data.textEmbeddings import TextEmbeddingLoader
from .data.cifar10 import Cifar10Loader


class DataLoader:
    @validate_call
    def __init__(self, data_name: TDataName = "mnist"):
        self.data_name = data_name
        self.loader = self._get_loader()

    def _get_loader(self):
        loader_mapping = {
            "mnist": lambda name: MnistSeriesLoader(name),
            "fmnist": lambda name: MnistSeriesLoader(name),
            "kmnist": lambda name: MnistSeriesLoader(name),
            "celegans": lambda name: CelegansLoader(),
            "ionosphere": lambda name: UcimlLoader(name),
            "optical_recognition": lambda name: UcimlLoader(name),
            "raisin": lambda name: UcimlLoader(name),
            "htru2": lambda name: UcimlLoader(name),
            "20ng": lambda name: TextEmbeddingLoader(name),
            "ag_news": lambda name: TextEmbeddingLoader(name),
            "amazon_polarity": lambda name: TextEmbeddingLoader(name),
            "yelp_review": lambda name: TextEmbeddingLoader(name),
            "cifar10": lambda name: Cifar10Loader(),
        }
        loader_provider = loader_mapping.get(self.data_name)
        if loader_provider is None:
            raise ValueError(f"Invalid type: {self.data_name}")
        return loader_provider(self.data_name)

    def get_data(self) -> DataModel:
        return self.loader.get_data()
